.NOLIST
.586
.model flat,stdcall
option casemap:none
include	\masm32\include\windows.inc

Rand32		proto	:DWORD
InitRand32	proto   :DWORD

.data?
seed	dd	?

.code
DllEntry	proc	hinstance:HINSTANCE, reason:DWORD, reserved1:DWORD
		invoke  InitRand32,0
		mov	eax,TRUE
		ret
DllEntry	endp

;Usage: RandomIntArray(Data:pointer,count:int,base:int,seed:int),int
;Returns a series of random integers, base>n>=0, to a dword array.  If seed=0, the
;CPU clock is used for random seeding.  Otherwise, a given seed will reproduce the
;same sequence.
RandomIntArray	proc uses esi pData:PTR DWORD, count:DWORD, base:DWORD, sede:dword
	invoke  InitRand32,sede
	mov     esi,pData
	mov     ecx,count
	mov     edx,base
@@:	or      ecx,ecx
	jz      @F
	push    ecx
	push    edx
	invoke  Rand32,edx
	pop     edx
	pop     ecx
	mov     [esi],eax
	add     esi,4
	dec     ecx
	jmp     @B
@@:	ret
RandomIntArray  endp

;Usage: InitRand32(seed:int)
;Sets a starting seed for random number generation.  Valid values range 1 
;through 7ffffffe.  An input value of zero uses the cpu time-stamp clock 
;as a 'randomize' option.  Values greater than 7ffffffe are ignored and 
;the seed is unchanged.  A negative seed avoids reseeding, allowing queries
;without changes.  The function returns the current seed and sets the Z flag 
;if this equals the requested input seed.
;
InitRand32	proc	new_seed:DWORD
	mov	eax,new_seed	;valid seeds range 1 thru 7ffffffeh
	test	eax,eax
	jnz	ir01		;If non-zero seed specified
@@:	rdtsc			; else get value from CPU clock
	and	eax,7fffffffh	; and clear msbit for a valid seed
	jz	@B		;IF TSC seed just happens to be 0 then retry

ir01:	cmp	eax,7fffffffh	;If seed > 7ffffffe, no changes
	jb	@F
	mov	eax,seed
@@:	mov	seed,eax
	cmp	eax,new_seed	;set flags returned for possible testing
	ret			;return seed value
InitRand32	endp

;IBM RANDU
;According to Numerical Recipes x = (x * 65539) mod (2^31)
;The parity of x (even/odd) never changes and that of (x mod base) never
;changes if base is even.  In practice, it appears that seeds were first
;scaled to floating point numbers, 0->1, and integers then obtained 
;from the most-significant bits of the result.  If nothing else, it's a 
;good example of why a prime number should be used as a seed modulus.
Rand32	proc	base:DWORD
	mov	eax,seed
	mov	ecx,65539
	mul	ecx
	mov	ecx,80000000h
	div	ecx
	mov	eax,edx
	mov	seed,eax
	xor	edx,edx
	div	base
	mov	eax,edx
	ret
Rand32	endp

End DllEntry
