;Statpack.asm - statistics library subroutines
;Copyright (c) 2005 Abel Sjurson
;Free software for distribution or modification subject to the terms of the GNU
;General Public License, v. 2.1, as published by the Free Software Foundation.

;Changes:
;02/12/05:  Modifications for InitRand32

.NOLIST
.586
.model flat,stdcall
option casemap:none
include	\masm32\include\windows.inc
include	\masm32\include\kernel32.inc
includelib \masm32\lib\kernel32.lib

LogGamma        proto	:PTR QWORD,:PTR QWORD
IncompleteGamma	proto	:PTR QWORD,:PTR QWORD,:PTR QWORD,:DWORD
Rand32		proto	:DWORD
InitRand32	proto   :DWORD
ChiSqProb	proto	:PTR QWORD,:PTR QWORD,:PTR QWORD,:DWORD

max_iterations	equ 1000

.data

;Lanczos g=9 coefficients given by Paul Godfrey
coef	dq	     1.000000000000000174663
	dq	  5716.400188274341379136
	dq	-14815.30426768413909044
	dq	 14291.49277657478554025
	dq	 -6348.160217641458813289
	dq	  1301.608286058321874105
	dq	  -108.1767053514369634679
	dq	     2.605696505611755827729
	dq	    -0.7423452510201416151527e-2
	dq	     0.5384136432509564062961e-7
	dq	    -0.4023533141268236372067e-8
gp5	dq	9.5

half	dq	0.5
dp1	dq	?
dp2	dq	?
seed	dd	?

.code
DllEntry	proc	hinstance:HINSTANCE, reason:DWORD, reserved1:DWORD
		invoke  InitRand32,0
		mov	eax,TRUE
		ret
DllEntry	endp

;Usage: GaussProb(x:pointer, P:pointer, flags:int)
;On input, x is a pointer to a 64-bit floating point number for which the
;Gaussian or normal probability function is to be evaluated.  If flags=0,
;P(x) is returned.  If flags=1, Q(x) is returned. ( P(x)+Q(x)=1 ).
;Return values:
;  2  Iteration maxed out

GaussProb	proc	uses esi edi	x:PTR QWORD, PQ:PTR QWORD, flags:DWORD
LOCAL	flgs:DWORD
LOCAL	y:QWORD

	finit
	mov	flgs,0

;y=x*x/2
	fld1
	fchs				;-1
	mov	esi,x
	fld	qword ptr [esi]		;x		-1
	fmul	st,st(0)		;x*x		-1
	fscale
	fstp	y
	fstp	st(0)

;res=IncompleteGamma(0.5,y,P,flags)
;result also returned to y
	invoke IncompleteGamma, ADDR half, ADDR y, ADDR y, flags
	mov	flgs,eax
	fld	y			;P or Q returned by call

	mov	eax, [esi+4]		;ms bytes of x
	mov	esi,PQ
	test	flags,1
	jnz	gpb01

;If P(x)
;If x>=0 then P(x)=0.5*(1+P) else P(x)=0.5*(1-P)
	or	eax,eax			;P
	jns	@F
	fchs				;-P
@@:	fld1				;1		P
	fadd	st(1),st		;1		1+P
	fchs				;-1		1+P
	fxch	st(1)			;(1+P)		-1
	fscale
	jmp	gpb02

;If Q(x)
;;If x>=0 then Q(x)=0.5*Q else Q(x)=1-0.5*Q
gpb01:	fld1
	fchs				;-1		Q
	fxch	st(1)			;Q		-1
	fscale				;0.5*Q		-1
	or	eax,eax
	jns	gpb02
	fadd	st,st(1)		;-1+0.5*Q	-1
	fchs				;1-0.5*Q	-1

gpb02:	fstp	qword ptr [esi]		;return result for P(x) or Q(x)
	fstp	st(0)			;clear stack

	mov	eax,flgs
	ret
GaussProb	endp

;Usage: ChiSqProb(nu:pointer, chi_sq:pointer, PQ:pointer, flags:int)
;On input, nu>0 and chi_sq>=0 are pointers to 64-bit floating point numbers for which the
;Chi-Square probabability is to be evaluated.  If flags=0, P(chisq,nu) is returned.
;If flags=1, Q(chisq,nu) is returned. ( P(chisq,nu)+Q(chisq,nu)=1 ).
;Return values:
;  1  Invalid z parameter (a<=0, x<0)
;  2  Iteration maxed out

ChiSqProb	proc	uses esi chi_sq:PTR QWORD, nu:PTR QWORD, PQ:PTR QWORD, flags:DWORD
LOCAL	x:QWORD
LOCAL	a:QWORD

	finit

;x=chi_sq/2:a=nu/2
	fld1
	fchs				;-1
	mov	esi,chi_sq
	fld	qword ptr [esi]		;chi_sq		-1
	fscale				;chi_sq/2	-1
	fstp	x
	mov	esi,nu
	fld	qword ptr [esi]		;nu		-1
	fscale				;nu/2		-1
	fstp	a
	fstp	st(0)

	invoke IncompleteGamma, ADDR a, ADDR x, PQ, flags
	ret
ChiSqProb	endp

;Usage: LogGamma(z:pointer, lngamma:pointer)
;On input, z is a pointer to a 64-bit floating point number (>0) whose gamma
;function is to be found.  lngamma is a pointer to a 64-bit floating point
;number to which the computed value of the natural logarithm of the gamma
;function will be returned.  Coding uses the Lanczos approximation for g=9 
;with coefficients given by Godfrey.
;Return values:
;  1  Invalid z parameter (<=0)

LogGamma	proc	uses esi edi	z:PTR QWORD, lngamma:PTR QWORD

	finit

;check that z>0
	mov	esi,z
	mov	eax,[esi+4]		;ms bytes of Z
	test	eax,eax
	ja	@F			;If >0
	mov	ax,1
	jmp	exit

;sum=c[0]:y=z
@@:	mov	edi,offset coef
	fld	qword ptr [edi]		;sum
	fld	qword ptr [esi]		;y		sum

;FOR i=1 TO 10:y=y+1.:sum=sum+c[i]/y:NEXT i
	mov	ecx,1			;ECX=I
@@:	fld1				;1		y	sum
	faddp	st(1),st		;y'		sum
	fld	qword ptr [edi+8*ecx]	;C[I]		y	sum
	fdiv	st,st(1)		;C[I]/y		y	sum
	faddp	st(2),st		;y		sum'
	inc	ecx
	cmp	ecx,10
	jbe	@B
	fstp	st(0)			;sum

;tmp=z+9.5:lngamma=(z+0.5)*LOG(tmp)-tmp+LOG(SQRT(2*pi)*sum/z))
	fdiv	qword ptr [esi]		;sum/z
	fldpi
	fadd	st,st(0)		;2*pi			sum/z
	fsqrt
	fmulp	st(1),st		;SQRT(2*PI)*sum/z
	fldln2
	fxch	st(1)			;SQRT(2*PI)*sum/z	ln(2)
	fyl2x				;LOG(SQRT(2*PI)*sum/z)

	fld	qword ptr [esi]		;z			LOG(..)
	fadd	gp5			;tmp=z+9.5		LOG(..)
	fldln2				;ln(2)			tmp	LOG(..)
	fld	st(1)			;tmp			ln(2)	tmp	LOG(..)
	fyl2x				;ln(tmp)		tmp	LOG(..)
	fld	qword ptr [esi]		;z			ln(tmp)	tmp	LOG(..)
	fadd	half			;z+0.5			ln(tmp)	tmp	LOG(..)
	fmulp	st(1),st		;(z+0.5)*ln(tmp)	tmp	LOG(..)
	faddp	st(2),st		;tmp			(z+0.5)*ln(tmp)+LOG(..)
	fsubp	st(1),st		;lngamma
	mov	esi,lngamma
	fstp	qword ptr [esi]
	mov	eax,0
exit:	ret

LogGamma	endp


;Usage: IncompleteGamma(a:pointer, x:pointer, P:pointer, flags:int)
;On input, a>0 and x>=0 are pointers to 64-bit floating point numbers for which the
;incomplete Gamma function is to be evaluated.  Two algorithms are available for
;determining the function, a series expansion and a continued fraction computation.
;By default the former is used when x<(a+1).  Flag bits may be used to override default
;behavior.  The effects of flag bits are:
;  Bit 0:  If set, Q(a,x) is returned, else P(a,x) is returned.
;  Bit 4:  If set, use of the series expansion, Pcalc is forced.
;  Bit 5:  If set, use of the continued fraction expansion, Qcalc is forced.
;
;Return values:
;  1  Invalid z parameter (a<=0, x<0)
;  2  Iteration maxed out

IncompleteGamma	proc	uses esi edi	a:PTR QWORD, x:PTR QWORD, PQ:PTR QWORD, flags:DWORD
LOCAL	lngamma:QWORD
LOCAL 	flgs:DWORD

	finit
	mov	flgs,0

;Check that a>0 and x>=0
	mov	eax,1			;error code
	mov	esi,a
	mov	edx,[esi+4]		;ms bytes of a
	test	edx,edx
	jbe	exit
	mov	edi,x
	mov	edx,[edi+4]		;ms bytes of x
	test	edx,edx
	jb	exit
	ja	@F

;If x=0 then P=0
	fldz
	xor	eax,eax
	jmp	igf00

;Find ln (Gamma(a))
@@:	invoke	LogGamma, a, ADDR lngamma
	fld	lngamma

	fld	qword ptr [esi]		;a		lngamma
	fld	qword ptr [edi]		;x		a	lngamma

;Test for overrides
	test	flags,10h		;Pcalc override
	jnz	igfpp
	test	flags,20h		;Qcalc override
	jnz	igfqq

;Default - If x<(a+1), calculate P(a,x) else calculate Q(a,x)
	fld1				;1		x	a	lngamma
	fadd	st,st(2)		;a+1		x	a	lngamma
	fcomp	st(1)			;x		a	lngamma
	fstsw	ax
	sahf
	ja	igfpp			;If a+1 > x

igfqq:	call	Qcalc
	test	flags,1
	jnz	igf01			;If Q(a,x) requested
	fld1				;1		Q(a,x)
	fsubrp	st(1),st		;P(a,x)
	jmp	igf01

igfpp:	call	Pcalc
igf00:	test	flags,1
	jz	igf01			;If P(a,x) requested
	fld1				;1		P(a,x)
	fsubrp	st(1),st		;Q(a,x)

igf01:	mov	flgs,eax
	mov	edi,PQ
	fstp	qword ptr [edi]		;Return value

	mov	eax,flgs
exit:	ret
IncompleteGamma	endp

;Pcalc uses a series expansion to compute P(a,x).
;  x, a and lngamma are input on the floating-point stack.
;Returns with P(a,x) on the stack and eax=2 for iteration failure.

Pcalc:
;b=a:sum=1./a:del=sum
	fld	st(1)			;b	x	a	lngamma
	fld1				;1	b	x	a	lngamma
	fdiv	st,st(3)		;sum	b	x	a	lngamma
	fld	st(0)			;del	sum	b	x	a	lngamma

;For i=0 to itermax
	mov	ecx,0			;ECX = iteration counter
	xor	edx,edx			;EDX = flag for iteration failure
@@:

;b=b+1.:del=del*x/b
	fld1				;1	del	sum	b	x	a	lngamma
	faddp	st(3),st		;del	sum	b'	x	a	lngamma
	fmul	st,st(3)		;del*x	sum	b	x	a	lngamma
	fdiv	st,st(2)		;del'	sum	b	x	a	lngamma

;If (sum+del)=sum then break else sum=sum+del
	fld	st(1)			;sum	del	sum	b	x	a	lngamma
	fadd	st,st(1)		;sum'	del	sum	b	x	a	lngamma
	fxch	st(2)			;sum	del	sum'	b	x	a	lngamma
	fcomp	st(2)			;del	sum'	b	x	a	lngamma
	fstsw	ax
	sahf
	jz	@F			;If sum=sum'

;NEXT i
	inc	ecx
	cmp	ecx,max_iterations
	jb	@B
	or	edx,2			;If failure to converge

;P=sum*EXP(-x+a*log(x)-lngamma)
@@:	fstp	st(0)			;sum		b	x		a	lngamma
	fldln2				;ln(2)		sum	b		x	a	lngamma
	fld	st(3)			;x		ln(2)	sum		b	x	a	lngamma
	fyl2x				;ln(x)		sum	b		x	a	lngamma
	fmulp	st(4),st		;sum		b	x		a*ln(x)	lngamma
	fxch	st(4)			;lngamma	b	x		a*ln(x)	sum
	fsubp	st(3),st		;b		x	a*ln(x)-lngamma	sum
	fstp	st(0)			;x		a*ln(x)-lngamma		sum
	fsubp	st(1),st		;a*ln(x)-lngamma-x	sum
	call	fexp			;exp(..)	sum
	fmulp	st(1),st		;P(a,x)
	mov	eax,edx			;iteration flag
	ret

;Qcalc uses a continued fraction expansion to compute Q(a,x).
;  x, a and lngamma are input on the floating-point stack.
;Returns with Q(a,x) on the stack and eax=2 for iteration failure.

Qcalc:
;To get some extra stack space store x and lngamma
	fst	dp1			;x		a	lngamma
	fxch	st(2)			;lngamma	a	x
	fstp	dp2			;a		x

;b=x+1-a:d=1./b:c=b+2:d=1./b:h=d
	fld1				;1		a	x
	fsub	st,st(1)		;1-a		a	x
	faddp	st(2),st		;a		b
	fld1				;1		a	b
	fdiv	st,st(2)		;d		a	b
	fld1				;1		d	a	b
	fadd	st,st(0)		;2		d	a	b
	fadd	st,st(3)		;c		d	a	b
	fld	st(1)			;h		c	d	a	b

;FOR i=1 TO 100
	mov	ecx,1			;ECX=I
	fld1				;i		h	c	d	a	b
	xor	edx,edx			;EDX = flag for iteration failure
					; Aso sets Z flag for first pass
qc01:

;a1=i*(a-i):b=b+2.:d=a1*d+b
	fld1				;1		i	h	c	d	a	b
	fadd	st(6),st
	faddp	st(6),st		;i		h	c	d	a	b+2
	fld	st(4)			;a		i	h	c	d	a	b+2
	fsub	st,st(1)		;a-i		i	h	c	d	a	b
	fmul	st,st(1)		;a1		i	h	c	d	a	b
	fxch	st(4)			;d		i	h	c	a1	a	b
	fmul	st,st(4)		;a1*d		i	h	c	a1	a	b
	fadd	st,st(6)		;d'		i	h	c	a1	a	b
	fxch	st(4)			;a1		i	h	c	d	a	b

;If i>1 then c=b+a1/c
	jz	@F			;skip on 1st pass
	fdiv	st,st(3)		;a1/c		i	h	c	d	a	b
	fadd	st,st(6)		;a1/c+b		i	h	c	d	a	b
	fst	st(3)			;a1/c+b		i	h	c	d	a	b
@@:	fstp	st(0)			;i		h	c	d	a	b

;d=1./d:del=c*d:h=h*del
	fld1				;1		i	h	c	d	a	b
	fdivrp	st(4),st		;i		h	c	d'	a	b
	fld	st(3)			;d		i	h	c	d'	a	b
	fmul	st,st(3)		;c*d		i	h	c	d'	a	b
	fmul	st(2),st		;del		i	h'	c	d'	a	b

;If del=1 then break
	fld1				;1		del	i	h'	c	d'	a	b
	fadd	st(2),st		;1		del	i+1	h'	c	d'	a	b
	fcompp				;i		h	c	d	a	b
	fstsw	ax
	sahf
	jz	@F

;Next i
	inc	ecx			;Note that i in the stack is already incremented
	cmp	ecx,max_iterations
	jb	qc01
	or	edx,2			;If failure to converge

;Clear the stack of all but a and h and recover x and lngamma
@@:	fstp	st(0)			;h		c	d	a	b
	fxch	st(4)			;b		c	d	a	h
	fstp	st(0)
	fstp	st(0)
	fstp	st(0)			;a		h
	fld	dp2			;lngamma	a	h
	fld	dp1			;x		lngamma	a	h

;Q=h*exp(-x+a*log(x)-lngamma)
	fldln2				;ln(2)		x		lngamma		a	h
	fld	st(1)			;x		ln(2)		x		lngamma	a	h
	fyl2x				;ln(x)		x		lngamma		a	h
	fmulp	st(3),st		;x		lngamma		a*ln(x)		h
	fsubp	st(2),st		;lngamma	a*ln(x)-x	h
	fsubp	st(1),st		;a*ln(x)-x-lngamma		h
	call	fexp			;exp(..)	h
	fmulp	st(1),st		;Q(a,x)
	mov	eax,edx			;iteration flag
	ret

;Subroutine to replace the value, x, at the top of the floating-point
; stack with exp(x).  Note that two free additional stack spaces are required.
fexp:	fldl2e				;log2(e)	x
	fmulp	st(1),st		;y=log2(e)*x
	fld	st(0)			;y		y
	frndint				;int(y)		y
	fsub	st(1),st		;int(y)		y-int(y)
	fxch				;y-int(y)	int(y)
	f2xm1				;2^(...)-1	int(y)
	fld1				;1		2^(...)-1	int(y)
	faddp	st(1),st		;2^(...)	int(y)
	fscale				;2^y		int(y)
	fstp	st(1)			;2^y
	ret

;Usage: RandomIntArray(Data:pointer,count:int,base:int,seed:int),int
;Returns a series of random integers, base>n>=0, to a dword array.  If seed=0, the
;CPU clock is used for random seeding.  Otherwise, a given seed will reproduce the
;same sequence.
RandomIntArray	proc uses esi pData:PTR DWORD, count:DWORD, base:DWORD, sede:dword
	invoke  InitRand32,sede
	mov     esi,pData
	mov     ecx,count
	mov     edx,base
@@:	or      ecx,ecx
	jz      @F
	push    ecx
	push    edx
	invoke  Rand32,edx
	pop     edx
	pop     ecx
	mov     [esi],eax
	add     esi,4
	dec     ecx
	jmp     @B
@@:	ret
RandomIntArray  endp

;Usage: DifferentiateArray(Data:pointer,count:int,base:int),int
;This function performs a circular differentiation of an array of N 32-bit 
;integers, i.e. Data[I]=Data[I+1]-Data[I].  If the difference is <0, base is 
;added so that all values are >=0 and <base.
DifferentiateArray	proc uses ebx esi pData:PTR DWORD, count:DWORD, base:DWORD

;FOR I=0 TO N-2:J=Data[I+1]-Data[I]:IF J<0 THEN J=J+base:Data[I]=J:NEXT J
	mov     esi,pData			;ESI -> Data[I]
	mov	ebx,[esi]			;EBX=Data[0]
	mov     ecx,count
	cmp	ecx,2
	jl	da02				;Return if N<2
	dec	ecx				;ECX=N-1 (>0)
	mov     edx,base
da01:	mov	eax,[esi+4]			;Data[I+1]
	sub	eax,[esi]
	jns	@F
	add	eax,edx
@@:	mov	[esi],eax			;Data[I]=Data[J]-Data[I]
	add     esi,4
	dec     ecx
	jnz     da01

;Data[N-1]=Data[0]-Data[N-1]
	mov	eax,ebx
	sub	eax,[esi]
	jns	@F
	add	eax,edx
@@:	mov	[esi],eax
da02:	ret
DifferentiateArray	endp

;Usage: IntegrateArray(Data:pointer,count:int,base:int),int
;This function is the inverse of DifferentiateArray and will regenerate the ;input array of the former apart from a constant of integration added to all ;values.
IntegrateArray	proc uses ebx esi pData:PTR DWORD, count:DWORD, base:DWORD

;B=Data[0]:Data[0]=0
	mov     esi,pData			;ESI -> Data[0]
	mov	ebx,[esi]			;EBX=B
	mov	dword ptr [esi],0
	add	esi,4				;ESI -> Data[1]
	mov     ecx,count
	cmp	ecx,2
	jl	ia02				;Return if N<2
	dec	ecx				;ECX=N-1 (>0)
	mov     edx,base

;FOR I=1 TO N-1:A=B:B=Data[I]:A=A+Data[I-1]
ia01:	mov	eax,ebx
	mov	ebx,[esi]
	add	eax,[esi-4]

;IF A>=base THEN A=A-base:Data[I]=A
	cmp	eax,edx
	jb	@F
	sub	eax,edx
@@:	mov	[esi],eax

;NEXT I
	add     esi,4
	dec     ecx
	jnz     ia01

ia02:	ret
IntegrateArray	endp

;Usage: RandomTest(data[0]:pointer, N:int, base:int, chisq:pointer, prob:pointer)
;On input, data[0] is a pointer to an array of N 32-bit integers having values
;0, 1, 2, ..., base-1.  The function returns the chi-square value assuming a 
;uniform box distribution for this range and its corresponding Q probability.
;Return values:
;  Bit 0:  If set, 1  invalid parameters (N<=1, base<=1)
;  Bit 1:  If set, Iteration maxed out
;  Bit 2:  If set, memory allocation failure
;  Bit 3:  If set, out of range values detected
RandomTest	proc	uses esi edi	data:PTR DWORD, N:DWORD, base:DWORD, chisq:PTR QWORD, prob:PTR QWORD
LOCAL	bins_:DWORD
LOCAL	trash:DWORD
LOCAL	nu:QWORD
LOCAL	flgs:DWORD

	finit
	mov	trash,0

;Test parameters - N>1, base>1
	mov	eax,1
	cmp	N,eax
	ja	@F
	or	flgs,1
	jmp	rt90
@@:	cmp	base,eax
	ja	@F
	or	flgs,1
	jmp	rt90

;Allocate memory for dword bins[0,1,...,base-1]
@@:	mov	eax,base
	shl	eax,2
	invoke	GlobalAlloc, GPTR, eax	;Allocate and zero bin memory
	test	eax,eax
	jnz	@F
	or	flgs,4
	jmp	rt90

@@:	mov	bins_,eax

;Fill bins with data counts
	mov	esi,data
	mov	edi,bins_
	mov	ecx,0
	mov	edx,base
rt01:	mov	eax,[esi+4*ecx]
	cmp	eax,edx
	jae	@F
	inc	dword ptr [edi+4*eax]
	jmp	rt02
@@:	inc	trash			;counter for bytes out of range
	or	flgs,8
rt02:	inc	ecx
	cmp	ecx,N
	jb	rt01

;n=N/base - expected mean count for random box distribution
	fild	N			;N
	fidiv	base			;n

;FOR i=0 to base-1:sum=sum+(n[i]-n)^2:NEXT i
	fldz				;sum		n
	mov	ecx,0			;ECX=I
@@:	fild	dword ptr [edi+4*ecx]	;n[i]		sum	n
	fsub	st,st(2)		;n[i]-n		sum	n
	fmul	st,st(0)		;(n[i]-n)^2	sum	n
	faddp	st(1),st		;sum'		n
	inc	ecx
	cmp	ecx,base
	jb	@B

;chisq=sum/n
	fdiv	st,st(1)		;chisq		n
	mov	edi,chisq
	fstp	qword ptr [edi]		;Return chi-square value

;nu=base-1				;degrees of freedom
	fld1				;1		n
	fild	base			;base		1	n
	fsub	st,st(1)		;nu		1	n
	fstp	nu
	fstp	st(0)			;clear stack
	fstp	st(0)

;find chisq probability Q(X2|nu)
	invoke	ChiSqProb, chisq, ADDR nu, prob, 1
	mov	flgs,eax

	invoke	GlobalFree, bins_
rt90:	mov	eax,flgs
	ret

RandomTest	endp

;*******************************************************************
;Insert any alternative RNG and seeder here.
;*******************************************************************
;Usage: InitRand32(seed:int)
;Sets a starting seed for random number generation.  Valid values range 1 
;through 7ffffffe.  An input value of zero uses the cpu time-stamp clock 
;as a 'randomize' option.  Values greater than 7ffffffe are ignored and 
;the seed is unchanged.  A negative seed avoids reseeding, allowing queries
;without changes.  The function returns the current seed and sets the Z flag 
;if this equals the requested input seed.
;
InitRand32	proc	new_seed:DWORD
	mov	eax,new_seed	;valid seeds range 1 thru 7ffffffeh
	test	eax,eax
	jnz	ir01		;If non-zero seed specified
@@:	rdtsc			; else get value from CPU clock
	and	eax,7fffffffh	; and clear msbit for a valid seed
	jz	@B		;IF TSC seed just happens to be 0 then retry

ir01:	cmp	eax,7fffffffh	;If seed > 7ffffffe, no changes
	jb	@F
	mov	eax,seed
@@:	mov	seed,eax
	cmp	eax,new_seed	;set flags returned for possible testing
	ret			;return seed value
InitRand32	endp

;Usage: Rand32(base:int)
;Returns a number ranging from 0 to base-1.  Valid base values range 1 
;through 7fffffff.  If the requested base lies outside this range, 
; -1 is returned, indicating the error.
;
Rand32	proc	base:DWORD
	mov	eax,base		;valid values 1 thru 7fffffff
	test	eax,eax
	jle	@F			;skip if base <= 0 (invalid)
	mov	eax,seed    
	mov	ecx,16807		;a = 7^5
	mul	ecx			;edx:eax == a*seed
	mov	ecx,7fffffffh
	div	ecx
	mov	eax,edx			;(a*seed) % 7fffffff
	mov	seed,eax		;save next seed 
	mov	ecx,base

;For divide modulus returns values 0 through base -1, 7ffffffeh max
;Favors lower seed bits
	mov	edx,0
	div	ecx

;Michael W's alternative: favors upper seed bits
;returns values 0 through 7ffffffdh max
;	shl	eax,1
;	mul	ecx

	mov	eax,edx
	ret

@@:	or	eax,-1			;if base error, return -1
	ret				; and set S flag.
Rand32	endp

End DllEntry
